"""Transfer LiquidHandleMethod

    :copyright: 2021 by The Autoprotocol Development Team, see AUTHORS
        for more details.
    :license: BSD, see LICENSE for more details

Base LiquidHandleMethod used by Protocol.transfer to generate a series of
movements between pairs of wells.
"""
from ..instruction import LiquidHandle
from ..unit import Unit
from ..util import parse_unit
from .liquid_handle_method import LiquidHandleMethod


# pylint: disable=unused-argument,too-many-instance-attributes,protected-access
class Transfer(LiquidHandleMethod):
    """LiquidHandleMethod for generating transfers between pairs of wells

    LiquidHandleMethod for transferring volume from one well to another.

    Attributes
    ----------
    _source_liquid : LiquidClass
        used to determine calibration, flowrates, and sensing thresholds
    _destination_liquid : LiquidClass
        used to determine calibration, flowrates, and sensing thresholds

    Notes
    -----
    The primary entry points that for this class are:
        - _aspirate_transports : generates transports for a source location
        - _dispense_transports : generates transports for a destination location


    See Also
    --------
    LiquidHandleMethod : base LiquidHandleMethod with reused functionality
    Protocol.transfer : the standard interface for interacting with Transfer
    """

    def __init__(
        self,
        tip_type=None,
        blowout=True,
        prime=True,
        transit=True,
        mix_before=False,
        mix_after=True,
        aspirate_z=None,
        dispense_z=None,
    ):
        """
        Parameters
        ----------
        tip_type : str, optional
            tip_type to be used for the LiquidHandlingMethod
        blowout : bool or dict, optional
            whether to execute a blowout step or the parameters for one.
            this generates a pair of operations: an initial air aspiration
            before entering any wells and a corresponding air dispense after the
            last operation that involves liquid
            See Also LiquidHandle.builders.blowout
        prime : bool or Unit, optional
            whether to execute a prime step or the parameters for one.
            this generates a pair of aspirate/dispense operations around the
            aspiration step in the sequence:
            aspirate_prime -> aspirate_target_volume -> dispense_prime
        transit : bool or Unit, optional
            whether to execute a transit step or the parameters for one.
            this generates a pair of operations wherein air is aspirated just
            before leaving the source location and dispensed immediately
            after reaching the destination location
        mix_before : bool or dict, optional
            whether to execute a mix_before step or the parameters for one.
            this generates a series of aspirate and dispense steps within the
            source location before aspirating the target volume
            See Also LiquidHandle.builders.mix
        mix_after : bool or dict, optional
            whether to execute a mix_after step or the parameters for one.
            this generates a series of aspirate and dispense steps within the
            destination location after dispensing the target volume
            See Also LiquidHandle.builders.mix
        aspirate_z : dict, optional
            the position that the tip should move to prior to aspirating, if the
            position references the `liquid_surface` then aspirate movements
            will track the surface with the defined offset.
            See Also LiquidHandle.builders.position_z
        dispense_z : dict, optional
            the position that the tip should move to prior to dispensing, if the
            position references the `liquid_surface` then dispense
            will track the surface with the defined offset.
            See Also LiquidHandle.builders.position_z
        """
        super(Transfer, self).__init__(tip_type=tip_type, blowout=blowout)

        # parameters for required behavior
        self.aspirate_z = aspirate_z
        self.dispense_z = dispense_z

        # parameters for optional behavior
        self.prime = prime
        self.transit = transit
        self.mix_before = mix_before
        self.mix_after = mix_after

        # LiquidHandle parameters that are generated and modified at runtime
        self._source_liquid = None
        self._destination_liquid = None

    def _has_calibration(self):
        liquids = [self._source_liquid, self._destination_liquid]
        return any(_ and _._has_calibration() for _ in liquids)

    def _calculate_overage_volume(self, volume):
        calibration_overage = (
            self._estimate_calibrated_volume(volume, self._source_liquid, self.tip_type)
            - volume
        )

        # handle whichever is larger, prime or transit volume
        if self.prime is True:
            prime = self.default_prime(volume)
        elif self.prime is False:
            prime = Unit(0, "uL")
        else:
            prime = self.prime

        if self.transit is True:
            transit = self.default_transit(volume)
        elif self.transit is False:
            transit = Unit(0, "uL")
        else:
            transit = self.transit

        prime_or_transit = max([prime, transit])

        return calibration_overage + prime_or_transit

    def default_blowout(self, volume):
        if self._is_single_channel():
            if volume < Unit("10:ul"):
                blowout_vol = Unit("5:ul")
            elif volume < Unit("25:ul"):
                blowout_vol = Unit("10:ul")
            elif volume < Unit("75:ul"):
                blowout_vol = Unit("15:ul")
            elif volume < Unit("100:ul"):
                blowout_vol = Unit("20:ul")
            else:
                blowout_vol = Unit("25:ul")
        else:
            blowout_vol = Unit("5:ul")

        return LiquidHandle.builders.blowout(
            volume=blowout_vol,
            initial_z=self.default_well_top_position_z(),
            flowrate=None,
        )

    def _aspirate_transports(self, volume, density):
        """Generates source well transports

        Generates and returns all of the transports that should happen within
        the source well of a transfer operation.

        Calls a series of _transport_`y` helper methods that each query the `y`
        parameter and default_`y` method to decide on a set of behavior and use
        that to define transports that are appended to the _transports list.

        Parameters
        ----------
        volume : Unit

        Return
        ------
        list
            source well transports corresponding to the aspirate operation

        Notes
        -----
        This method defines what lower level transport-generating methods are
        called and in what order. It can be overwritten when adding an
        entirely new set of transport-generating behavior.

        See Also
        --------
        _dispense_transports : corresponding destination well method
        """
        self._transports = []
        volume = parse_unit(Unit(volume), "ul")

        # No transports if no volume specified
        if volume == Unit("0:ul"):
            return []

        self._transport_pre_buffer(volume)
        self._transport_mix_before(volume)
        self._transport_aspirate_target_volume(volume, density)
        self._transport_aspirate_transit(volume)

        return self._transports

    def _dispense_transports(self, volume, density):
        """Generates destination well transports

        Generates and returns all of the transports that should happen within
        the destination well of a transfer operation.

        Calls a series of _transport_`y` helper methods that each query the `y`
        parameter and default_`y` method to decide on a set of behavior and use
        that to define transports that are appended to the _transports list.

        Parameters
        ----------
        volume : Unit

        Return
        ------
        list
            destination well transports corresponding to the dispense operation

        Notes
        -----
        This method defines what lower level transport-generating methods are
        called and in what order. It can be overwritten when adding an
        entirely new set of transport-generating behavior.

        See Also
        --------
        _aspirate_transports : corresponding source well method
        """
        self._transports = []
        volume = parse_unit(volume, "ul")

        # No transports if no volume specified
        if volume == Unit("0:ul"):
            return []

        self._transport_dispense_transit(volume)
        self._transport_dispense_target_volume(volume, density)
        self._transport_mix_after(volume)
        self._transport_blowout(volume)

        return self._transports

    def _transport_mix_before(self, volume):
        """Mixes volume in the source well before aspirating

        Parameters
        ----------
        volume : Unit

        See Also
        --------
        mix_before : holds any user defined mix_before parameters
        default_mix_before : specifies default mix_before parameters
        _mix : lower level helper that generates the mix_before transports
        """
        if self.mix_before is True:
            mix_before = self.default_mix_before(volume)
        elif self.mix_before is False:
            mix_before = False
        else:
            mix_before = self.mix_before

        if mix_before is not False:
            mix_before = LiquidHandle.builders.mix(**mix_before)
            self._mix(
                delay_time=self._source_liquid.delay_time,
                liquid_class=self._source_liquid.name,
                **mix_before
            )

    # pragma pylint: disable=missing-param-doc
    def default_mix_before(self, volume: Unit):
        """Default mix_before parameters

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        dict
            mix_before params

        See Also
        --------
        mix_before : holds any user defined mix_before parameters
        _transport_mix : generates the actual mix_before transports
        """
        if self._is_single_channel():
            mix_z = self.default_lld_position_z(liquid=self._source_liquid)
        else:
            mix_z = self.default_well_bottom_position_z()

        return LiquidHandle.builders.mix(
            volume=volume,
            repetitions=10,
            initial_z=mix_z,
            asp_flowrate=self._source_liquid._get_aspirate_flowrate(
                volume, self.tip_type
            ),
            dsp_flowrate=self._source_liquid._get_dispense_flowrate(
                volume, self.tip_type
            ),
        )

    def _transport_aspirate_target_volume(self, volume, density):
        """Aspirates the target volume from the source location

        Parameters
        ----------
        volume : Unit
        density : Unit

        See Also
        --------
        aspirate_z : holds any user defined aspirate_z parameters
        default_aspirate_z : specifies default aspirate_z parameters
        prime : holds any user defined prime volume
        default_prime : specifies default prime volume
        _aspirate_simple : lower level helper that generates aspirate transports
        _aspirate_with_prime : lower level helper for aspirating with priming
        """
        aspirate_z = self.aspirate_z or self.default_aspirate_z(volume)

        if self.prime is True:
            prime = self.default_prime(volume)
        elif self.prime is False:
            prime = False
        else:
            prime = self.prime

        aspirate_z = LiquidHandle.builders.position_z(**aspirate_z)
        if prime is not False:
            prime = parse_unit(prime, "uL")
            self._aspirate_with_prime(
                volume=volume,
                prime_vol=prime,
                calibrated_vol=self._source_liquid._get_calibrated_volume(
                    volume, self.tip_type
                ),
                initial_z=aspirate_z,
                asp_flowrate=self._source_liquid._get_aspirate_flowrate(
                    volume, self.tip_type
                ),
                dsp_flowrate=self._source_liquid._get_dispense_flowrate(
                    volume, self.tip_type
                ),
                delay_time=self._source_liquid.delay_time,
                liquid_class=self._source_liquid.name,
                density=density,
            )
        else:
            self._aspirate_simple(
                volume=volume,
                calibrated_vol=self._source_liquid._get_calibrated_volume(
                    volume, self.tip_type
                ),
                initial_z=aspirate_z,
                flowrate=self._source_liquid._get_aspirate_flowrate(
                    volume, self.tip_type
                ),
                delay_time=self._source_liquid.delay_time,
                liquid_class=self._source_liquid.name,
                density=density,
            )

    # pragma pylint: disable=missing-param-doc
    def default_aspirate_z(self, volume: Unit):
        """Default aspirate_z parameters

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        dict
            aspirate position_z

        See Also
        --------
        aspirate_z : holds any user defined aspirate_z parameters
        _transport_aspirate_target_volume : generates actual aspirate transports
        """
        if self._is_single_channel():
            aspirate_z = self.default_lld_position_z(liquid=self._source_liquid)
        else:
            aspirate_z = self.default_well_bottom_position_z()
        return aspirate_z

    # pylint: disable=no-self-use, missing-param-doc
    def default_prime(self, volume: Unit):
        """Default prime volume

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        Unit
            priming volume

        See Also
        --------
        prime : holds any user defined prime volume
        _transport_aspirate_target_volume : generates actual aspirate transports
        """
        return Unit(5, "ul")

    def _transport_aspirate_transit(self, volume):
        """Aspirates air above the source before moving to the destination

        Parameters
        ----------
        volume : Unit

        See Also
        --------
        transit : holds any user defined transit volume
        default_transit : specifies default transit volume
        _transport_dispense_transit : the corresponding air dispense step
        """
        if self.transit is True:
            transit = self.default_transit(volume)
        elif self.transit is False:
            transit = False
        else:
            transit = self.transit

        if transit is not False:
            transit_vol = parse_unit(transit, "uL")
            self._aspirate_simple(
                volume=transit_vol,
                initial_z=self.default_well_top_position_z(),
                liquid_class="air",
            )

    def _transport_dispense_transit(self, volume):
        """Dispenses air above the destination after moving from the source

        Parameters
        ----------
        volume : Unit

        See Also
        --------
        transit : holds any user defined transit volume
        default_transit : specifies default transit volume
        _transport_aspirate_transit : the corresponding air aspirate step
        """
        if self.transit is True:
            transit = self.default_transit(volume)
        elif self.transit is False:
            transit = False
        else:
            transit = self.transit

        if transit is not False:
            transit_vol = parse_unit(transit, "uL")
            self._dispense_simple(
                volume=transit_vol,
                initial_z=self.default_well_top_position_z(),
                liquid_class="air",
            )

    # pragma pylint: disable=missing-param-doc
    def default_transit(self, volume: Unit):
        """Default transit volume

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        Unit
            transit volume

        See Also
        --------
        transit : holds any user defined transit volume
        _transport_aspirate_transit : generates the actual transit transports
        _transport_dispense_transit : generates the actual transit transports
        """
        if self._is_single_channel():
            transit_vol = Unit("2:ul")
        else:
            transit_vol = Unit("1:ul")

        return transit_vol

    def _transport_dispense_target_volume(self, volume, density):
        """Dispenses the target volume into the destination location

        Parameters
        ----------
        volume : Unit
        density : Unit

        See Also
        --------
        dispense_z : holds any user defined dispense_z parameters
        default_dispense_z : specifies default dispense_z parameters
        _dispense_simple : lower level helper that generates dispense transports
        """
        dispense_z = self.dispense_z or self.default_dispense_z(volume)

        dispense_z = LiquidHandle.builders.position_z(**dispense_z)
        self._dispense_simple(
            volume=volume,
            calibrated_vol=self._source_liquid._get_calibrated_volume(
                volume, self.tip_type
            ),
            initial_z=dispense_z,
            flowrate=self._source_liquid._get_dispense_flowrate(volume, self.tip_type),
            delay_time=self._source_liquid.delay_time,
            liquid_class=self._source_liquid.name,
            density=density,
        )

    # pragma pylint: disable=missing-param-doc
    def default_dispense_z(self, volume: Unit):
        """Default aspirate_z parameters

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        dict
            dispense position_z

        See Also
        --------
        dispense_z : holds any user defined dispense_z parameters
        _transport_dispense_target_volume : generates actual dispense transports
        """
        if self._is_single_channel():
            dispense_z = self.default_lld_position_z(liquid=self._destination_liquid)
        else:
            dispense_z = self.default_tracked_position_z()

        return dispense_z

    def _transport_mix_after(self, volume):
        """Mixes volume in the destination well after dispensing

        Parameters
        ----------
        volume : Unit

        See Also
        --------
        mix_after : holds any user defined mix_after parameters
        default_mix_after : specifies default mix_after parameters
        _mix : lower level helper that generates the mix_after transports
        """
        if self.mix_after is True:
            mix_after = self.default_mix_after(volume)
        elif self.mix_after is False:
            mix_after = False
        else:
            mix_after = self.mix_after

        if mix_after is not False:
            mix_after = LiquidHandle.builders.mix(**mix_after)
            self._mix(
                delay_time=self._source_liquid.delay_time,
                liquid_class=self._source_liquid.name,
                **mix_after
            )

    # pragma pylint: disable=missing-param-doc
    def default_mix_after(self, volume: Unit):
        """Default mix_after parameters

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        dict
            mix_after params

        See Also
        --------
        mix_after : holds any user defined mix_after parameters
        _transport_mix : generates the actual mix_after transports
        """
        if self._is_single_channel():
            mix_z = self.default_lld_position_z(liquid=self._destination_liquid)
        else:
            mix_z = self.default_well_bottom_position_z()

        return LiquidHandle.builders.mix(
            volume=volume,
            repetitions=10,
            initial_z=mix_z,
            asp_flowrate=self._source_liquid._get_aspirate_flowrate(
                volume, self.tip_type
            ),
            dsp_flowrate=self._source_liquid._get_dispense_flowrate(
                volume, self.tip_type
            ),
        )


class DryWellTransfer(Transfer):
    """Dispenses while tracking liquid without mix_after"""

    def __init__(
        self,
        tip_type=None,
        blowout=True,
        prime=True,
        transit=True,
        mix_before=False,
        mix_after=False,
        aspirate_z=None,
        dispense_z=None,
    ):
        super(DryWellTransfer, self).__init__(
            tip_type=tip_type,
            blowout=blowout,
            prime=prime,
            transit=transit,
            mix_before=mix_before,
            mix_after=mix_after,
            aspirate_z=aspirate_z,
            dispense_z=dispense_z,
        )

    def default_dispense_z(self, volume):
        return self.default_tracked_position_z()


class PreMixBlowoutTransfer(Transfer):
    """Adds an additional blowout before the mix_after step"""

    def __init__(
        self,
        tip_type=None,
        blowout=True,
        prime=True,
        transit=True,
        mix_before=False,
        mix_after=True,
        aspirate_z=None,
        dispense_z=None,
        pre_mix_blowout=True,
    ):
        super(PreMixBlowoutTransfer, self).__init__(
            tip_type=tip_type,
            blowout=blowout,
            prime=prime,
            transit=transit,
            mix_before=mix_before,
            mix_after=mix_after,
            aspirate_z=aspirate_z,
            dispense_z=dispense_z,
        )
        self.pre_mix_blowout = pre_mix_blowout

    def _dispense_transports(self, volume=None, density=None):
        self._transports = []
        volume = parse_unit(volume, "ul")

        # No transports if no volume specified
        if volume == Unit("0:ul"):
            return []

        self._transport_dispense_transit(volume)
        self._transport_dispense_target_volume(volume, density)
        self._transport_pre_mix_blowout(volume)
        self._transport_mix_after(volume)
        self._transport_blowout(volume)

        return self._transports

    def _calculate_pre_buffer(self, volume):
        if self.blowout is True:
            blowout = self.default_blowout(volume)
        elif self.blowout is False:
            blowout = {}
        else:
            blowout = self.blowout

        if self.pre_mix_blowout is True:
            secondary_blowout = self.default_pre_mix_blowout(volume)
        elif self.pre_mix_blowout is False:
            secondary_blowout = {}
        else:
            secondary_blowout = self.pre_mix_blowout

        blowout_vol = parse_unit(blowout.get("volume", Unit("0:uL")), "uL")
        secondary_blowout_vol = parse_unit(
            secondary_blowout.get("volume", Unit("0:uL")), "uL"
        )

        return blowout_vol + secondary_blowout_vol

    def _transport_pre_mix_blowout(self, volume):
        """Dispenses a secondary air volume befiore the mix_after step

        Notes
        -----
        For some liquid classes this has resulted in more complete dispensing of
        the target volume than just a single blowout.

        Parameters
        ----------
        volume : Unit

        See Also
        --------
        pre_mix_blowout : holds any user defined pre_mix_blowout parameters
        default_pre_mix_blowout : specifies default pre_mix_blowout parameters
        """
        if self.pre_mix_blowout is True:
            pre_mix_blowout = self.default_pre_mix_blowout(volume)
        elif self.pre_mix_blowout is False:
            pre_mix_blowout = False
        else:
            pre_mix_blowout = self.pre_mix_blowout

        if pre_mix_blowout is not False:
            pre_mix_blowout = LiquidHandle.builders.blowout(**pre_mix_blowout)
            self._dispense_simple(liquid_class="air", **pre_mix_blowout)

    # pragma pylint: disable=missing-param-doc
    def default_pre_mix_blowout(self, volume: Unit):
        """Default pre_mix_blowout parameters

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        dict
            pre_mix_blowout params

        See Also
        --------
        pre_mix_blowout : holds any user defined pre_mix_blowout parameters
        _transport_pre_mix_blowout : generates the actual blowout transports
        """
        return LiquidHandle.builders.blowout(
            volume=Unit(5, "ul"),
            initial_z=self.default_well_top_position_z(),
            flowrate=None,
        )



from dataclasses import dataclass
from typing import List, Union

from ..container import Well, WellGroup


WellParam = Union[Well, List[Well], WellGroup]


@dataclass
class AutopickGroup:
    source: WellParam
    destination: WellParam
    min_abort: int = 0



import json


class TestUtils:
    @staticmethod
    def read_json_file(file_path: str):
        file = open("./test/data/{0}".format(file_path), encoding="utf-8")
        data = json.load(file)
        return json.dumps(data, indent=2, sort_keys=True)



"""LiquidHandleMethod

Base class for generating complex liquid handling behavior.

    :copyright: 2021 by The Autoprotocol Development Team, see AUTHORS
        for more details.
    :license: BSD, see LICENSE for more details

Summary
-------
LiquidHandleMethods are passed as arguments to Protocol methods along with
LiquidClasses to specify complex series of liquid handling behaviors.

Notes
-----
Methods in this file should not be used directly, but are intended to be
extended by other methods depending on desired behavior.

When creating a vendor-specific library it's likely desirable to monkey patch
`LiquidHandleMethod._get_tip_types` to reference TipTypes that the vendor
supports.
"""
from ..instruction import LiquidHandle
from ..unit import Unit
from ..util import parse_unit
from .tip_type import TipType


# pylint: disable=too-many-public-methods,protected-access
class LiquidHandleMethod(object):
    """Base LiquidHandleMethod

    General framework for liquid handling abstractions and helpers for
    building a series of liquid_handle transports.

    Attributes
    ----------
    _shape : dict
        the SBS shape and number of rows and columns of the liquid_handle
    _transports : list
        tracks transports to be added to the LiquidHandle instruction

    Notes
    -----
    There is a hierarchy of logic to all LiquidHandleMethods that abstracts a
    complex set of liquid handling behavior into smaller, discrete steps.

    For step `x` (aspirate, dispense, mix) and parameter `y` (e.g. blowout):
        - Protocol method:
            - calls LiquidHandleMethod._`x`_transports
        - LiquidHandleMethod._`x`_transports method:
            - clears the _transports list
            - walks through all _transport methods including _transport_`y`
            - returns the _transports lists
        - LiquidHandleMethod._transport_`y` method:
            - checks parameter `y` in addition to the default_`y` method
            - possibly generates a series of transports based on the two values
            - calls lower level helper methods
        - LiquidHandleMethod lower level helper methods:
            - generate transports and append them to _transports

    Examples
    --------
    For specifying a single, global liquid handling behavior across all
    volumes the easiest way is to specify parameters when instantiating a
    LiquidHandleMethod.

    .. code-block:: python

        from autoprotocol import Unit
        from autoprotocol.instruction import LiquidHandle
        from autoprotocol.liquid_handle import LiquidHandleMethod

        lhm = LiquidHandleMethod(
            blowout=LiquidHandle.builders.blowout(volume=Unit(10, "uL"))
        )

    For behavior that relies on more liquid handling parameters or even defines
    new behavior you can define your own LiquidHandleMethod.

    .. code-block:: python

        from autoprotocol import Unit
        from autoprotocol.instruction import LiquidHandle
        from autoprotocol.liquid_handle import LiquidHandleMethod

        class NewLHM(LiquidHandleMethod):
            def default_blowout(self, volume):
                if volume < Unit(10, "uL"):
                    blowout_volume = Unit(1, "uL")
                else:
                    blowout_volume = Unit(10, "uL")
                return LiquidHandle.builders.blowout(
                    volume=blowout_volume
                )

    See Also
    --------
    Transfer : method for handling liquid between two locations
    Mix : method for handling liquid within locations
    LiquidClass : contain properties that are intrinsic to specific liquids
    Protocol : contains methods that accept LiquidHandleMethods as arguments
    """

    def __init__(self, tip_type=None, blowout=True):
        """
        Parameters
        ----------
        tip_type : str, optional
            tip_type to be used for the LiquidHandlingMethod
        blowout : bool or dict, optional
            whether to execute a blowout step or the parameters for one.
            this generates two operations, an initial air aspiration before
            entering any wells, and a corresponding final air dispense
            after the last operation that involves liquid
            See Also LiquidHandle.builders.blowout
        """
        self.tip_type = tip_type
        self.blowout = blowout

        # LiquidHandle parameters that are generated and modified at runtime
        self._shape = None
        self._transports = []

    def _get_tip_types(self):
        """Gets a list of TipTypes based on _shape

        Decides on the TipTypes that are compatible with the _shape of the
        liquid handling operation.

        Returns
        -------
        list
            list of valid TipTypes

        Notes
        -----
        This method exists to be monkey patched in vendor libraries to map
        shape parameters to actual tip types.

        Raises
        ------
        RuntimeError
            if there are no tips that support the _shape

        See Also
        --------
        _get_sorted_tip_types : the standard interface for this method
        """
        if self._is_single_channel():
            tip_types = [
                TipType("generic_1_50", Unit("50:ul")),
                TipType("generic_1_1000", Unit("1000:ul")),
            ]
        elif self._shape["format"] == "SBS96":
            tip_types = [TipType("generic_96_180", Unit("180:ul"))]
        elif self._shape["format"] == "SBS384":
            tip_types = [TipType("generic_384_30", Unit("30:ul"))]
        else:
            raise RuntimeError(f"No tip types supported for shape: {self._shape}")

        return tip_types

    def _get_sorted_tip_types(self):
        """Gets a list of valid TipTypes in ascending order of volume

        Returns
        -------
        list
            list of valid TipTypes sorted by their maximum capacity

        See Also
        --------
        _get_tip_types : vendor library-specific tip selection method
        """
        return sorted(self._get_tip_types(), key=lambda t: t.volume)

    def _rec_tip_type(self, volume):
        """For a given volume gets the smallest appropriate tip type

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        Unit
            the recommended tip type for a given volume

        Raises
        ------
        RuntimeError
            if there is no tip large enough for the specified volume
        """
        total_vol = volume + self._calculate_overage_volume(volume)
        tips = self._get_sorted_tip_types()
        valid_tips = list(filter(lambda t: t.volume >= total_vol, tips))
        if not valid_tips:
            raise RuntimeError(
                f"None of the TipTypes: {tips} allowed for this shape are "
                f"large enough to hold: {total_vol} (the target volume + "
                f"overage)."
            )
        return valid_tips[0].name

    def _tip_capacity(self):
        """Gets the best estimate of tip capacity with the given information

        Uses either the defined or a calculated tip_type to estimate the
        total usable volume for a given tip after accounting for overage volume.

        Returns
        -------
        Unit
            the maximum allowable volume for the method's tip_type

        See Also
        --------
        _calculate_overage_volume : calculates volume unusable for liquid
        """
        tip_types = self._get_sorted_tip_types()
        if self.tip_type is None:
            tip_capacity = tip_types[-1].volume
        else:
            tips_by_name = {_.name: _ for _ in tip_types}
            tip_capacity = tips_by_name[self.tip_type].volume

        overage_vol = self._calculate_overage_volume(tip_capacity).ceil()
        max_volume = tip_capacity - overage_vol

        return max_volume

    def _has_calibration(self):
        """Checks whether any specified LiquidClasses specify any calibrations

        Checks whether any of the LiquidClasses used by the
        LiquidHandleMethod have any volume or flowrate calibration.

        Returns
        -------
        bool
            whether there will be any calibration based on LiquidClasses

        See Also
        --------
        LiquidClass._has_calibration : checks for calibration parameters
        """
        raise NotImplementedError

    @staticmethod
    def _estimate_calibrated_volume(volume, liquid, tip_type):
        """Gives an estimation of calibrated volume

        If tip_type is specified then gets the actual calibrated volume,
        but if it is None then gets a rough estimate of the maximum volume
        calibration that might be experienced.

        This is used for estimates of calibrated volume used for calculating
        overage volume within the tip.

        Parameters
        ----------
        volume : Unit
            the uncalibrated volume
        liquid : LiquidClass
            the liquid class to be calibrated against
        tip_type : str or None
            the name of the TipType to be used

        Returns
        -------
        Unit
            an estimate of the calibrated volume

        See Also
        --------
        _calculate_overage_volume : uses this method
        LiquidClass._safe_volume_multiplier : used if no tip_type is specified
        """

        estimated = liquid._safe_volume_multiplier * volume
        if tip_type:
            calculated = liquid._get_calibrated_volume(volume, tip_type)
            cal_vol = calculated if calculated is not None else estimated
        else:
            cal_vol = estimated

        return cal_vol

    def _calculate_overage_volume(self, volume):
        """Calculates extra volume held in the tip besides the target volume

        Calculates how much extra volume is contained within the tip besides
        the target volume. This includes things like overage due to
        calibrated volumes being larger than the nominal volume.

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        Unit
            the extra volume taken up in the tip besides the transfer volume
        """
        raise NotImplementedError

    def _is_single_channel(self):
        """Determines whether _shape represents a single channel

        Returns
        -------
        bool
            whether or not the method's shape is single channel-compatible
        """
        return self._shape["rows"] == 1 and self._shape["columns"] == 1

    def _aspirate_simple(
        self,
        volume,
        initial_z,
        position_x=None,
        position_y=None,
        calibrated_vol=None,
        flowrate=None,
        delay_time=None,
        liquid_class=None,
        density=None,
    ):
        """Helper function for generating aspirate transports

        Parameters
        ----------
        volume : Unit
            volume of liquid to be aspirated
        initial_z : dict
            position that the tip will move to before pump movement
        position_x : dict, optional
            position that the tip will move to before pump movement
        position_y : dict, optional
            position that the tip will move to before pump movement
        calibrated_vol : Unit, optional
            calibrated volume, volume which the pump will move
        flowrate : dict, optional
            flowrate of liquid during aspiration
        delay_time : Unit, optional
            time to pause after aspirating to let pressure equilibrate
        liquid_class : str, optional
            the name of the liquid class being aspirated
        density : Unit, optional
            the density of liquid being aspirated
        """

        followup_z = self._move_to_initial_position(position_x, position_y, initial_z)

        mode_params = LiquidHandle.builders.mode_params(
            position_x=position_x,
            position_y=position_y,
            position_z=followup_z,
            liquid_class=liquid_class,
        )

        self._transports += [
            LiquidHandle.builders.transport(
                volume=-volume,
                density=density,
                # pylint: disable=invalid-unary-operand-type
                pump_override_volume=-calibrated_vol if calibrated_vol else None,
                flowrate=flowrate,
                mode_params=mode_params,
                delay_time=delay_time,
            )
        ]

    def _aspirate_with_prime(
        self,
        volume,
        prime_vol,
        initial_z,
        position_x=None,
        position_y=None,
        calibrated_vol=None,
        asp_flowrate=None,
        dsp_flowrate=None,
        delay_time=None,
        liquid_class=None,
        density=None,
    ):
        """Helper function for generating aspiration with priming

        Parameters
        ----------
        volume : Unit
            volume of liquid to be aspirated
        prime_vol : Unit
            volume of additional liquid to be aspirated along with volume
        initial_z : dict
            position that the tip will move to before pump movement
        position_x : dict, optional
            position that the tip will move to before pump movement
        position_y : dict, optional
            position that the tip will move to before pump movement
        calibrated_vol : Unit, optional
            calibrated volume, volume which the pump will move
        asp_flowrate : dict, optional
            flowrate of liquid during aspiration
        dsp_flowrate : dict, optional
            flowrate of liquid during aspiration
        delay_time : Unit, optional
            time to pause after aspirating to let pressure equilibrate
        liquid_class : str, optional
            the name of the liquid class being aspirated
        density : Unit, optional
            the density of liquid being aspirated
        """

        followup_z = self._move_to_initial_position(position_x, position_y, initial_z)

        mode_params = LiquidHandle.builders.mode_params(
            position_x=position_x,
            position_y=position_y,
            position_z=followup_z,
            liquid_class=liquid_class,
        )

        # Aspirate with priming volume
        self._transports += [
            LiquidHandle.builders.transport(
                volume=-prime_vol,
                density=density,
                flowrate=asp_flowrate,
                mode_params=mode_params,
                delay_time=delay_time,
            ),
            LiquidHandle.builders.transport(
                volume=-volume,
                density=density,
                # pylint: disable=invalid-unary-operand-type
                pump_override_volume=(-calibrated_vol if calibrated_vol else None),
                flowrate=asp_flowrate,
                mode_params=mode_params,
                delay_time=delay_time,
            ),
            LiquidHandle.builders.transport(
                volume=prime_vol,
                density=density,
                flowrate=dsp_flowrate,
                mode_params=mode_params,
                delay_time=delay_time,
            ),
        ]

    def _dispense_simple(
        self,
        volume,
        initial_z,
        position_x=None,
        position_y=None,
        calibrated_vol=None,
        flowrate=None,
        delay_time=None,
        liquid_class=None,
        density=None,
    ):
        """Helper function for generating dispense transports

        Parameters
        ----------
        volume : Unit
            volume of liquid to be dispensed
        initial_z : dict
            position that the tip will move to before pump movement
        position_x : dict, optional
            position that the tip will move to before pump movement
        position_y : dict, optional
            position that the tip will move to before pump movement
        calibrated_vol : Unit, optional
            calibrated volume, volume which the pump will move
        flowrate : dict, optional
            flowrate of liquid during dispense
        delay_time : Unit, optional
            time to pause after dispensing to let pressure equilibrate
        liquid_class : str, optional
            the name of the liquid class being dispensed
        density : Unit, optional
            the density of liquid to be dispensed
        """

        followup_z = self._move_to_initial_position(position_x, position_y, initial_z)

        mode_params = LiquidHandle.builders.mode_params(
            position_x=position_x,
            position_y=position_y,
            position_z=followup_z,
            liquid_class=liquid_class,
        )

        self._transports += [
            LiquidHandle.builders.transport(
                volume=volume,
                density=density,
                pump_override_volume=calibrated_vol if calibrated_vol else None,
                flowrate=flowrate,
                mode_params=mode_params,
                delay_time=delay_time,
            )
        ]

    def _mix(
        self,
        volume,
        repetitions,
        position_x=None,
        position_y=None,
        initial_z=None,
        asp_flowrate=None,
        dsp_flowrate=None,
        delay_time=None,
        liquid_class=None,
    ):
        """Helper function for generating mix transports

        Parameters
        ----------
        volume : Unit
            volume of liquid to be aspirated and expelled during mixing
        repetitions : int
            number of times to aspirate and expel liquid during mixing
        initial_z : dict
            position_z of the tip during the move_before_mix transport that
            happens before mixing
        position_x : dict, optional
            position that the tip will move to before pump movement
        position_y : dict, optional
            position that the tip will move to before pump movement
        asp_flowrate : dict, optional
            flowrate of liquid aspiration during mixing
        dsp_flowrate : dict, optional
            flowrate of liquid dispensing during mixing
        delay_time : Unit, optional
            time to pause after dispensing to let pressure equilibrate
        liquid_class : str, optional
            the name of the liquid class being dispensed
        """

        followup_z = self._move_to_initial_position(position_x, position_y, initial_z)

        mode_params = LiquidHandle.builders.mode_params(
            position_x=position_x,
            position_y=position_y,
            position_z=followup_z,
            liquid_class=liquid_class,
        )

        self._transports += [
            LiquidHandle.builders.transport(
                volume=-volume,
                flowrate=asp_flowrate,
                mode_params=mode_params,
                delay_time=delay_time,
            ),
            LiquidHandle.builders.transport(
                volume=volume,
                flowrate=dsp_flowrate,
                mode_params=mode_params,
                delay_time=delay_time,
            ),
        ] * repetitions

    def _move_to_initial_position(
        self, position_x=None, position_y=None, position_z=None
    ):
        """Moves to a given position_z and then returns a followup one

        Takes an initial position_z and moves to it before returning a
        suitable followup.

        If sensing is specified, then moves to well top and senses before
        returning a suitable followup position.

        Parameters
        ----------
        position_x : dict, optional
            initial x position to move to before aspirating/dispensing
        position_y : dict, optional
            initial y position to move to before aspirating/dispensing
        position_z : dict, optional
            initial z position to move to before aspirating/dispensing

        Returns
        -------
        dict
            followup z position to pipette at after initial movement

        See Also
        --------
        _move_to_well_top_before_lld : helper method
        _get_followup_z : helper method
        """
        self._move_to_well_top_before_lld(position_z)

        if any([position_z, position_x, position_y]):
            self._transports += [
                LiquidHandle.builders.transport(
                    mode_params=LiquidHandle.builders.mode_params(
                        position_x=position_x,
                        position_y=position_y,
                        position_z=position_z,
                    )
                )
            ]

        followup_z = self._get_followup_z(position_z)

        return followup_z

    def _move_to_well_top_before_lld(self, position_z):
        """If position_z contains any liquid sensing moves to well top

        Parameters
        ----------
        position_z : dict
            position_z to be checked for lld events
        """
        well_top_z = LiquidHandle.builders.transport(
            mode_params=LiquidHandle.builders.mode_params(
                position_z=LiquidHandle.builders.position_z(reference="well_top")
            )
        )

        if position_z:
            if position_z.get("reference") == "liquid_surface":
                if position_z.get("detection", {}).get("method") != "tracked":
                    self._transports.append(well_top_z)

    @staticmethod
    def _get_followup_z(position_z):
        """Generates a position_z that references preceding position/tracked

        Generates a position_z to followup after the specified one. If the
        liquid surface is referenced, then returns a position that similarly
        tracks the surface. Otherwise, generates a reference to the preceding
        position.

        Parameters
        ----------
        position_z : dict
            position to be referenced when deciding on a followup position

        Returns
        -------
        dict
            position_z to follow up at a non-sensing position
        """
        preceding_z = LiquidHandle.builders.position_z(reference="preceding_position")
        if not position_z:
            followup_z = preceding_z
        elif position_z["reference"] == "liquid_surface":
            followup_z = LiquidHandle.builders.position_z(
                reference="liquid_surface",
                offset=position_z.get("offset"),
                detection_method="tracked",
            )
        else:
            followup_z = preceding_z

        return followup_z

    def _transport_pre_buffer(self, volume):
        """Aspirates a pre buffer of air volume above the source location

        Parameters
        ----------
        volume : Unit

        See Also
        --------
        _calculate_pre_buffer : determines pre_buffer volume to aspirate
        _transport_blowout : the corresponding air dispense step
        """
        pre_buffer = self._calculate_pre_buffer(volume)
        if pre_buffer:
            pre_buffer = parse_unit(pre_buffer, "uL")
            self._aspirate_simple(
                volume=pre_buffer,
                initial_z=LiquidHandle.builders.position_z(
                    reference="well_top", offset="1:mm"
                ),
                liquid_class="air",
            )

    def _calculate_pre_buffer(self, volume):
        """Calculates a recommended pre_buffer volume

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        Unit
            pre_buffer volume

        See Also
        --------
        _transport_pre_buffer : generates the actual pre_buffer transports
        """
        if self.blowout is True:
            blowout = self.default_blowout(volume)
        elif self.blowout is False:
            blowout = {}
        else:
            blowout = self.blowout

        return blowout.get("volume", Unit("0:uL"))

    def _transport_blowout(self, volume):
        """Blows out air volume above the destination location

        Parameters
        ----------
        volume : Unit
            liquid handling volume

        See Also
        --------
        blowout : holds any user specified blowout parameters
        default_blowout : specifies default blowout parameters
        _transport_pre_buffer : the corresponding air aspirate step
        """
        if self.blowout is True:
            blowout = self.default_blowout(volume)
        elif self.blowout is False:
            blowout = False
        else:
            blowout = self.blowout

        if blowout is not False:
            blowout_params = LiquidHandle.builders.blowout(**blowout)
            self._dispense_simple(liquid_class="air", **blowout_params)

    # pylint: disable=missing-param-doc
    def default_blowout(self, volume: Unit):
        """Default blowout behavior

        Parameters
        ----------
        volume : Unit

        Returns
        -------
        dict
            blowout_params

        See Also
        --------
        blowout : holds any user specified blowout parameters
        _transport_blowout : generates the actual blowout transports
        """
        raise NotImplementedError

    @staticmethod
    def default_lld_position_z(liquid):
        """Default lld position_z

        Returns
        -------
        dict
            position_z for sensing the liquid surface
        """
        # TODO: Select thresholds in some order
        return LiquidHandle.builders.position_z(
            reference="liquid_surface",
            offset="-1:mm",
            detection_method="capacitance",
            detection_threshold=liquid.clld_threshold,
        )

    @staticmethod
    def default_tracked_position_z():
        """Default tracked position_z

        Returns
        -------
        dict
            position_z for tracking the liquid surface
        """
        return LiquidHandle.builders.position_z(
            reference="liquid_surface", detection_method="tracked", offset="-1:mm"
        )

    @staticmethod
    def default_well_bottom_position_z():
        """Default well bottom position_z

        Returns
        -------
        dict
            position_z for the well bottom
        """
        return LiquidHandle.builders.position_z(reference="well_bottom")

    @staticmethod
    def default_well_top_position_z():
        """Default well top position_z

        Returns
        -------
        dict
            position_z for the well top
        """
        return LiquidHandle.builders.position_z(reference="well_top")
